from pyautd3.native_methods.autd3 import GainSTMMode

from .control_point import ControlPoint, ControlPoints
from .foci import FociSTM
from .gain import GainSTM, GainSTMOption

__all__ = [
    "ControlPoint",
    "ControlPoints",
    "FociSTM",
    "GainSTM",
    "GainSTMMode",
    "GainSTMOption",
]



import ctypes
from collections.abc import Iterable
from typing import Self

import numpy as np
from numpy.typing import ArrayLike

from pyautd3.driver.common.freq import Freq
from pyautd3.driver.datagram.datagram import Datagram
from pyautd3.driver.datagram.stm.control_point import (
    ControlPoint,
    ControlPoints,
    ControlPoints1,
    ControlPoints2,
    ControlPoints3,
    ControlPoints4,
    ControlPoints5,
    ControlPoints6,
    ControlPoints7,
    ControlPoints8,
)
from pyautd3.driver.datagram.stm.stm_sampling_config import FreqNearest, PeriodNearest, _sampling_config
from pyautd3.driver.datagram.with_loop_behavior import DatagramL
from pyautd3.driver.datagram.with_segment import DatagramS
from pyautd3.driver.firmware.fpga.sampling_config import SamplingConfig
from pyautd3.driver.firmware.fpga.transition_mode import TransitionMode
from pyautd3.driver.geometry import Geometry
from pyautd3.native_methods.autd3 import Segment
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import DatagramPtr, FociSTMPtr, LoopBehavior, TransitionModeWrap
from pyautd3.utils import Duration


class FociSTM(DatagramS[FociSTMPtr], DatagramL[FociSTMPtr], Datagram):
    foci: list[ControlPoints]
    config: SamplingConfig | Freq[float] | Duration | FreqNearest | PeriodNearest

    @classmethod
    def __private_new__(
        cls: type["FociSTM"],
        foci: (Iterable[ArrayLike] | Iterable[ControlPoint] | Iterable[ControlPoints]),
        config: SamplingConfig | Freq[float] | Duration | FreqNearest | PeriodNearest,
    ) -> "FociSTM":
        ins = super().__new__(cls)
        ins.__private_init__(foci, config)
        return ins

    def __private_init__(
        self: "FociSTM",
        foci: (Iterable[ArrayLike] | Iterable[ControlPoint] | Iterable[ControlPoints]),
        config: SamplingConfig | Freq[float] | Duration | FreqNearest | PeriodNearest,
    ) -> None:
        foci_ = list(foci)
        match foci_[0]:
            case ControlPoints():
                self.foci = foci_  # type: ignore[assignment]
            case ControlPoint():
                self.foci = [ControlPoints(points=[p]) for p in foci_]  # type: ignore[list-item, arg-type]
            case _:
                self.foci = [ControlPoints(points=[ControlPoint(point=p)]) for p in foci_]  # type: ignore[arg-type]

        self.config = config

    def __init__(
        self: "FociSTM",
        foci: (Iterable[ArrayLike] | Iterable[ControlPoint] | Iterable[ControlPoints]),
        config: SamplingConfig | Freq[float] | Duration,
    ) -> None:
        self.__private_init__(foci, config)

    def into_nearest(self: Self) -> "FociSTM":
        match self.config:
            case Freq() as freq:
                return FociSTM.__private_new__(self.foci, FreqNearest(freq))  # type: ignore[arg-type]
            case Duration() as period:
                return FociSTM.__private_new__(self.foci, PeriodNearest(period))  # type: ignore[arg-type]
            case _:
                raise TypeError

    def _raw_ptr(self: Self, _: Geometry) -> FociSTMPtr:
        n = self._n()
        match n:
            case 1:
                foci = (ControlPoints1(p.points[0], p.intensity.value) for p in self.foci)
            case 2:
                foci = (ControlPoints2(p.points[0], p.points[1], p.intensity.value) for p in self.foci)  # type: ignore[misc]
            case 3:
                foci = (ControlPoints3(p.points[0], p.points[1], p.points[2], p.intensity.value) for p in self.foci)  # type: ignore[misc]
            case 4:
                foci = (ControlPoints4(p.points[0], p.points[1], p.points[2], p.points[3], p.intensity.value) for p in self.foci)  # type: ignore[misc]
            case 5:
                foci = (ControlPoints5(p.points[0], p.points[1], p.points[2], p.points[3], p.points[4], p.intensity.value) for p in self.foci)  # type: ignore[misc]
            case 6:
                foci = (
                    ControlPoints6(p.points[0], p.points[1], p.points[2], p.points[3], p.points[4], p.points[5], p.intensity.value)  # type: ignore[misc]
                    for p in self.foci
                )
            case 7:
                foci = (
                    ControlPoints7(p.points[0], p.points[1], p.points[2], p.points[3], p.points[4], p.points[5], p.points[6], p.intensity.value)  # type: ignore[misc]
                    for p in self.foci
                )
            case _:
                foci = (
                    ControlPoints8(  # type: ignore[misc]
                        p.points[0],
                        p.points[1],
                        p.points[2],
                        p.points[3],
                        p.points[4],
                        p.points[5],
                        p.points[6],
                        p.points[7],
                        p.intensity.value,
                    )
                    for p in self.foci
                )
        foci_ = np.fromiter((np.void(p) for p in foci), dtype=np.dtype((np.void, 4 + n * 16)))  # type: ignore[type-var,call-overload]
        return Base().stm_foci(
            self.sampling_config()._inner,
            foci_.ctypes.data_as(ctypes.c_void_p),  # type: ignore[arg-type]
            len(self.foci),
            n,
        )

    def _into_segment(self: Self, ptr: FociSTMPtr, segment: Segment, transition_mode: TransitionModeWrap | None) -> DatagramPtr:
        return Base().stm_foci_into_datagram_with_segment(ptr, self._n(), segment, transition_mode or TransitionMode.NONE)

    def _into_loop_behavior(
        self: Self,
        ptr: FociSTMPtr,
        segment: Segment,
        transition_mode: TransitionModeWrap | None,
        loop_behavior: LoopBehavior,
    ) -> DatagramPtr:
        return Base().stm_foci_into_datagram_with_loop_behavior(
            ptr,
            self._n(),
            segment,
            transition_mode or TransitionMode.NONE,
            loop_behavior,
        )

    def _datagram_ptr(self: Self, geometry: Geometry) -> DatagramPtr:
        return Base().stm_foci_into_datagram(self._raw_ptr(geometry), self._n())

    def sampling_config(self: Self) -> SamplingConfig:
        return _sampling_config(self.config, len(self.foci))

    def _n(self: Self) -> int:
        n = len(self.foci[0].points)
        if any(len(f.points) != n for f in self.foci):
            msg = "All components must have the same number of foci"
            raise ValueError(msg)
        match n:
            case v if 0 < v <= 8:  # noqa: PLR2004
                return n
            case _:
                raise ValueError



from typing import Self

from pyautd3.driver.utils import _validate_u8
from pyautd3.native_methods.autd3 import Intensity as Intensity_


class Intensity:
    value: int

    def __init__(self: Self, intensity: int) -> None:
        self.value = _validate_u8(intensity)

    def __floordiv__(self: Self, other: int) -> "Intensity":
        return Intensity(self.value // other)

    def __eq__(self: Self, __value: object, /) -> bool:
        return isinstance(__value, Intensity) and self.value == __value.value

    def __hash__(self: Self) -> int:
        return self.value.__hash__()  # pragma: no cover

    def __str__(self: Self) -> str:
        return f"Intensity({self.value})"

    def __repr__(self: Self) -> str:
        return f"Intensity({self.value})"

    def _inner(self: Self) -> Intensity_:
        return Intensity_(self.value)

    MIN: "Intensity" = None  # type: ignore[assignment]
    MAX: "Intensity" = None  # type: ignore[assignment]


Intensity.MIN = Intensity(0x00)
Intensity.MAX = Intensity(0xFF)



import numpy as np
import pytest

from pyautd3 import AUTD3, Clear, Controller, ForceFan, Segment
from pyautd3.autd_error import AUTDError, InvalidDatagramTypeError
from pyautd3.controller import FixedDelay, FixedSchedule
from pyautd3.controller.controller import SenderOption
from pyautd3.controller.sleeper import SpinSleeper, SpinStrategy, SpinWaitSleeper, StdSleeper
from pyautd3.driver.datagram import Synchronize
from pyautd3.driver.firmware.fpga.emit_intensity import Intensity
from pyautd3.driver.firmware.fpga.phase import Phase
from pyautd3.driver.firmware_version import FirmwareInfo
from pyautd3.gain import Uniform
from pyautd3.link.audit import Audit
from pyautd3.modulation import Static
from pyautd3.native_methods.autd3 import ParallelMode
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.utils.duration import Duration


def create_controller() -> Controller[Audit]:
    return Controller.open(
        [AUTD3(pos=[0.0, 0.0, 0.0], rot=[1.0, 0.0, 0.0, 0.0]), AUTD3(pos=[0.0, 0.0, 0.0], rot=[1.0, 0.0, 0.0, 0.0])],
        Audit(),
    )


def test_sender_is_default():
    assert Base().sender_option_is_default(SenderOption()._inner())


def test_sleeper():
    _ = StdSleeper()._inner()
    _ = SpinSleeper().with_spin_strategy(SpinStrategy.SpinLoopHint)._inner()
    _ = SpinWaitSleeper()._inner()


def test_firmware_info():
    autd: Controller[Audit]
    with create_controller() as autd:
        assert FirmwareInfo.latest_version() == "v12.0.0"

        for i, firm in enumerate(autd.firmware_version()):
            assert firm.info == f"{i}: CPU = v12.0.0, FPGA = v12.0.0 [Emulator]"
            assert str(firm) == f"{i}: CPU = v12.0.0, FPGA = v12.0.0 [Emulator]"


def test_close():
    autd: Controller[Audit]
    with create_controller() as autd:
        assert autd.link().is_open()

        autd.close()
        autd.close()

    with create_controller() as autd:
        autd.link().break_down()
        with pytest.raises(AUTDError) as e:
            autd.close()
        assert str(e.value) == "broken"


def test_sender():
    autd: Controller[Audit]
    with create_controller() as autd:
        for dev in autd.geometry():
            assert np.all(autd.link().modulation_buffer(dev.idx(), Segment.S0) == 0xFF)

        autd.sender(SenderOption(), FixedSchedule(StdSleeper())).send(Static(intensity=0x80))
        for dev in autd.geometry():
            assert np.all(autd.link().modulation_buffer(dev.idx(), Segment.S0) == 0x80)

        autd.sender(SenderOption(), FixedDelay(SpinWaitSleeper())).send(Static(intensity=0x81))
        for dev in autd.geometry():
            assert np.all(autd.link().modulation_buffer(dev.idx(), Segment.S0) == 0x81)


def test_send_single():
    autd: Controller[Audit]
    with create_controller() as autd:
        for dev in autd.geometry():
            assert np.all(autd.link().modulation_buffer(dev.idx(), Segment.S0) == 0xFF)

        autd.send(Static())

        for dev in autd.geometry():
            assert np.all(autd.link().modulation_buffer(dev.idx(), Segment.S0) == 0xFF)

        autd.link().break_down()
        with pytest.raises(AUTDError) as e:
            autd.send(Static())
        assert str(e.value) == "broken"
        autd.link().repair()


def test_send_tuple():
    autd: Controller[Audit]
    with create_controller() as autd:
        for dev in autd.geometry():
            assert np.all(autd.link().modulation_buffer(dev.idx(), Segment.S0) == 0xFF)
            intensities, phases = autd.link().drives_at(dev.idx(), Segment.S0, 0)
            assert np.all(intensities == 0)
            assert np.all(phases == 0)

        autd.send((Static(), Uniform(intensity=Intensity(0x80), phase=Phase(0))))
        for dev in autd.geometry():
            assert np.all(autd.link().modulation_buffer(dev.idx(), Segment.S0) == 0xFF)
            intensities, phases = autd.link().drives_at(dev.idx(), Segment.S0, 0)
            assert np.all(intensities == 0x80)
            assert np.all(phases == 0)

        with pytest.raises(InvalidDatagramTypeError):
            autd.send(0)  # type: ignore[arg-type]

        autd.link().break_down()
        with pytest.raises(AUTDError) as e:
            autd.send((Static(), Uniform(intensity=Intensity(0xFF), phase=Phase(0))))
        assert str(e.value) == "broken"
        autd.link().repair()


def test_clear():
    autd: Controller[Audit]
    with create_controller() as autd:
        autd.send((Static(), Uniform(intensity=Intensity(0xFF), phase=Phase(0x90))))

        for dev in autd.geometry():
            assert np.all(autd.link().modulation_buffer(dev.idx(), Segment.S0) == 0xFF)
            intensities, phases = autd.link().drives_at(dev.idx(), Segment.S0, 0)
            assert np.all(intensities == 0xFF)
            assert np.all(phases == 0x90)

        autd.send(Clear())

        for dev in autd.geometry():
            assert np.all(autd.link().modulation_buffer(dev.idx(), Segment.S0) == 0xFF)
            intensities, phases = autd.link().drives_at(dev.idx(), Segment.S0, 0)
            assert np.all(intensities == 0)
            assert np.all(phases == 0)


def test_synchronize():
    autd: Controller[Audit]
    with create_controller() as autd:
        autd.send(Synchronize())


def test_force_fan():
    autd: Controller[Audit]
    with create_controller() as autd:
        for dev in autd.geometry():
            assert not autd.link().is_force_fan(dev.idx())

        autd.send(ForceFan(lambda dev: dev.idx() == 0))
        assert autd.link().is_force_fan(0)
        assert not autd.link().is_force_fan(1)

        autd.send(ForceFan(lambda dev: dev.idx() == 1))
        assert not autd.link().is_force_fan(0)
        assert autd.link().is_force_fan(1)


def test_geometry():
    autd: Controller[Audit]
    with create_controller() as autd:
        assert autd.num_transducers() == 249 * 2
        assert autd[0].num_transducers() == 249


def test_sender_option():
    autd: Controller[Audit]
    with create_controller() as autd:
        option = SenderOption(
            send_interval=Duration.from_millis(1),
            receive_interval=Duration.from_millis(1),
            timeout=Duration.from_millis(100),
            parallel=ParallelMode.Off,
        )
        autd.default_sender_option = option
        assert option == autd.default_sender_option



from typing import Self

import numpy as np
from numpy.typing import ArrayLike

from pyautd3.driver.datagram.gain import Gain
from pyautd3.driver.firmware.fpga.emit_intensity import Intensity
from pyautd3.driver.firmware.fpga.phase import Phase
from pyautd3.driver.geometry import Geometry
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi import PlaneOption as PlaneOption_
from pyautd3.native_methods.autd3capi_driver import GainPtr
from pyautd3.native_methods.structs import Vector3


class PlaneOption:
    intensity: Intensity
    phase_offset: Phase

    def __init__(self: Self, *, intensity: Intensity | None = None, phase_offset: Phase | None = None) -> None:
        self.intensity = intensity or Intensity.MAX
        self.phase_offset = phase_offset or Phase(0)

    def _inner(self: Self) -> PlaneOption_:
        return PlaneOption_(self.intensity._inner(), self.phase_offset._inner())


class Plane(Gain):
    direction: np.ndarray
    option: PlaneOption

    def __init__(self: Self, direction: ArrayLike, option: PlaneOption) -> None:
        super().__init__()
        self.direction = np.array(direction)
        self.option = option

    def _gain_ptr(self: Self, _: Geometry) -> GainPtr:
        return Base().gain_plane(Vector3(self.direction), self.option._inner())



import ctypes
from collections.abc import Callable, Iterator
from typing import Self

import numpy as np

from pyautd3.driver.autd3_device import AUTD3
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import GeometryPtr
from pyautd3.native_methods.structs import Point3, Quaternion

from .device import Device


class Geometry:
    _geometry_ptr: GeometryPtr
    _devices: list[Device]

    def __init__(self: Self, ptr: GeometryPtr) -> None:
        self._geometry_ptr = ptr
        self._devices = [Device(i, ptr) for i in range(int(Base().geometry_num_devices(self._geometry_ptr)))]

    def center(self: Self) -> np.ndarray:
        return Base().geometr_center(self._geometry_ptr).ndarray()

    def num_devices(self: Self) -> int:
        return int(Base().geometry_num_devices(self._geometry_ptr))

    def num_transducers(self: Self) -> int:
        return int(Base().geometry_num_transducers(self._geometry_ptr))

    def __getitem__(self: Self, key: int) -> Device:
        return self._devices[key]

    def __iter__(self: Self) -> Iterator[Device]:
        return iter(self._devices)

    def set_sound_speed_from_temp(
        self: Self,
        temp: float,
        k: float = 1.4,
        r: float = 8.31446261815324,
        m: float = 28.9647e-3,
    ) -> None:
        for d in self:
            d.set_sound_speed_from_temp(temp, k, r, m)

    def set_sound_speed(
        self: Self,
        c: float,
    ) -> None:
        for d in self:
            d.sound_speed = c

    def reconfigure(self: Self, f: Callable[[Device], AUTD3]) -> None:
        devices = [f(d) for d in self._devices]
        pos = np.fromiter((np.void(Point3(d.pos)) for d in devices), dtype=Point3)  # type: ignore[type-var,call-overload]
        rot = np.fromiter((np.void(Quaternion(d.rot)) for d in devices), dtype=Quaternion)  # type: ignore[type-var,call-overload]
        Base().geometry_reconfigure(
            self._geometry_ptr,
            pos.ctypes.data_as(ctypes.POINTER(Point3)),  # type: ignore[arg-type]
            rot.ctypes.data_as(ctypes.POINTER(Quaternion)),  # type: ignore[arg-type]
        )
        self._devices = [Device(i, self._geometry_ptr) for i in range(int(Base().geometry_num_devices(self._geometry_ptr)))]



from collections.abc import Callable

from .controller import Controller, FixedSchedule, SenderOption
from .controller.sleeper import SpinSleeper
from .driver.autd3_device import AUTD3
from .driver.common import Hz, deg, kHz, rad
from .driver.datagram import (
    Clear,
    FixedCompletionTime,
    FixedUpdateRate,
    FociSTM,
    ForceFan,
    GainSTM,
    GainSTMOption,
    GPIOOutputs,
    GPIOOutputType,
    Group,
    PhaseCorrection,
    PulseWidthEncoder,
    ReadsFPGAState,
    Silencer,
    SwapSegment,
    WithLoopBehavior,
    WithSegment,
)
from .driver.datagram.stm import ControlPoint, ControlPoints
from .driver.firmware.fpga import Drive, Intensity, LoopBehavior, Phase, PulseWidth, SamplingConfig, TransitionMode
from .driver.geometry import Device, EulerAngles, Geometry, Transducer
from .ethercat import DcSysTime
from .gain import Bessel, BesselOption, Focus, FocusOption, GainGroup, Null, Plane, PlaneOption, Uniform
from .link.nop import Nop
from .modulation import Sine, SineOption, Square, SquareOption, Static
from .native_methods.autd3 import GainSTMMode, GPIOIn, GPIOOut, ParallelMode, Segment
from .utils import Duration

_ext_tracing_init: list[Callable[[], None]] = []


__all__ = [
    "AUTD3",
    "Bessel",
    "BesselOption",
    "Clear",
    "ControlPoint",
    "ControlPoints",
    "Controller",
    "Custom",
    "DcSysTime",
    "Device",
    "Drive",
    "Duration",
    "EulerAngles",
    "FixedCompletionTime",
    "FixedSchedule",
    "FixedUpdateRate",
    "FociSTM",
    "Focus",
    "FocusOption",
    "ForceFan",
    "GPIOIn",
    "GPIOOut",
    "GPIOOutputType",
    "GPIOOutputs",
    "GainGroup",
    "GainSTM",
    "GainSTMMode",
    "GainSTMOption",
    "Geometry",
    "Group",
    "Hz",
    "Intensity",
    "LoopBehavior",
    "Nop",
    "Null",
    "ParallelMode",
    "Phase",
    "PhaseCorrection",
    "Plane",
    "PlaneOption",
    "PulseWidth",
    "PulseWidthEncoder",
    "ReadsFPGAState",
    "SamplingConfig",
    "SamplingMode",
    "Segment",
    "SenderOption",
    "Silencer",
    "Sine",
    "SineOption",
    "SpinSleeper",
    "Square",
    "SquareOption",
    "Static",
    "SwapSegment",
    "Transducer",
    "TransitionMode",
    "Uniform",
    "WithLoopBehavior",
    "WithSegment",
    "deg",
    "kHz",
    "phase_rad",
    "rad",
    "tracing_init",
]

__version__ = "34.0.0"



from typing import Self

from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.utils import _validate_u16


class PulseWidth:
    _pulse_width: int

    @classmethod
    def __private_new__(cls: type["PulseWidth"], pulse_width: int) -> "PulseWidth":
        ins = super().__new__(cls)
        ins._pulse_width = pulse_width
        return ins

    def __init__(self: Self, pulse_width: int) -> None:
        self._pulse_width = _validate_u16(Base().pulse_width_512(pulse_width))

    @staticmethod
    def from_duty(duty: float) -> "PulseWidth":
        return PulseWidth.__private_new__(_validate_u16(Base().pulse_width_512_from_duty(duty)))

    @property
    def pulse_width(self: Self) -> int:
        return self._pulse_width



from typing import Self

from pyautd3.controller.sleeper import SpinSleeper, SpinWaitSleeper, StdSleeper
from pyautd3.native_methods.autd3capi_driver import TimerStrategyTag, TimerStrategyWrap


class FixedSchedule:
    _sleeper: StdSleeper | SpinSleeper | SpinWaitSleeper

    def __init__(self: Self, sleeper: StdSleeper | SpinSleeper | SpinWaitSleeper | None = None) -> None:
        if sleeper is None:
            sleeper = SpinSleeper()
        self._sleeper = sleeper

    def _inner(self: Self) -> TimerStrategyWrap:
        return TimerStrategyWrap(TimerStrategyTag.FixedSchedule, self._sleeper._inner())


class FixedDelay:
    _sleeper: StdSleeper | SpinSleeper | SpinWaitSleeper

    def __init__(self: Self, sleeper: StdSleeper | SpinSleeper | SpinWaitSleeper) -> None:
        self._sleeper = sleeper

    def _inner(self: Self) -> TimerStrategyWrap:
        return TimerStrategyWrap(TimerStrategyTag.FixedDelay, self._sleeper._inner())



from .controller import Controller, SenderOption
from .sleeper import SpinSleeper, SpinWaitSleeper, StdSleeper
from .strategy import FixedDelay, FixedSchedule

__all__ = ["Controller", "FixedDelay", "FixedSchedule", "SenderOption", "SpinSleeper", "SpinWaitSleeper", "StdSleeper", "WaitableSleeper"]



import ctypes
from typing import Self

import numpy as np
from numpy.typing import ArrayLike

from pyautd3.driver.firmware.fpga.emit_intensity import Intensity
from pyautd3.driver.firmware.fpga.phase import Phase
from pyautd3.native_methods.structs import Point3

# _pad is not required for python 3.12+?


class ControlPoint(ctypes.Structure):
    _fields_ = [  # noqa: RUF012
        ("_point", Point3),
        ("_offset", ctypes.c_uint8),
        ("_pad", ctypes.c_uint8 * 3),
    ]

    def __init__(self: Self, point: ArrayLike, phase_offset: Phase | None = None) -> None:
        super().__init__()
        self.point = np.array(point)
        self.phase_offset = phase_offset or Phase(0)

    @property
    def point(self: Self) -> np.ndarray:
        return self._point.ndarray()

    @point.setter
    def point(self: Self, value: ArrayLike) -> None:
        self._point = Point3(np.array(value))

    @property
    def phase_offset(self: Self) -> Phase:
        return Phase(self._offset)

    @phase_offset.setter
    def phase_offset(self: Self, value: Phase) -> None:
        self._offset = value.value


class ControlPoints1(ctypes.Structure):
    _fields_ = [  # noqa: RUF012
        ("_point", ControlPoint),
        ("_intensity", ctypes.c_uint8),
        ("_pad", ctypes.c_uint8 * 3),
    ]


class ControlPoints2(ctypes.Structure):
    _fields_ = [  # noqa: RUF012
        ("_point1", ControlPoint),
        ("_point2", ControlPoint),
        ("_intensity", ctypes.c_uint8),
        ("_pad", ctypes.c_uint8 * 3),
    ]


class ControlPoints3(ctypes.Structure):
    _fields_ = [  # noqa: RUF012
        ("_point1", ControlPoint),
        ("_point2", ControlPoint),
        ("_point3", ControlPoint),
        ("_intensity", ctypes.c_uint8),
        ("_pad", ctypes.c_uint8 * 3),
    ]


class ControlPoints4(ctypes.Structure):
    _fields_ = [  # noqa: RUF012
        ("_point1", ControlPoint),
        ("_point2", ControlPoint),
        ("_point3", ControlPoint),
        ("_point4", ControlPoint),
        ("_intensity", ctypes.c_uint8),
        ("_pad", ctypes.c_uint8 * 3),
    ]


class ControlPoints5(ctypes.Structure):
    _fields_ = [  # noqa: RUF012
        ("_point1", ControlPoint),
        ("_point2", ControlPoint),
        ("_point3", ControlPoint),
        ("_point4", ControlPoint),
        ("_point5", ControlPoint),
        ("_intensity", ctypes.c_uint8),
        ("_pad", ctypes.c_uint8 * 3),
    ]


class ControlPoints6(ctypes.Structure):
    _fields_ = [  # noqa: RUF012
        ("_point1", ControlPoint),
        ("_point2", ControlPoint),
        ("_point3", ControlPoint),
        ("_point4", ControlPoint),
        ("_point5", ControlPoint),
        ("_point6", ControlPoint),
        ("_intensity", ctypes.c_uint8),
        ("_pad", ctypes.c_uint8 * 3),
    ]


class ControlPoints7(ctypes.Structure):
    _fields_ = [  # noqa: RUF012
        ("_point1", ControlPoint),
        ("_point2", ControlPoint),
        ("_point3", ControlPoint),
        ("_point4", ControlPoint),
        ("_point5", ControlPoint),
        ("_point6", ControlPoint),
        ("_point7", ControlPoint),
        ("_intensity", ctypes.c_uint8),
        ("_pad", ctypes.c_uint8 * 3),
    ]


class ControlPoints8(ctypes.Structure):
    _fields_ = [  # noqa: RUF012
        ("_point1", ControlPoint),
        ("_point2", ControlPoint),
        ("_point3", ControlPoint),
        ("_point4", ControlPoint),
        ("_point5", ControlPoint),
        ("_point6", ControlPoint),
        ("_point7", ControlPoint),
        ("_point8", ControlPoint),
        ("_intensity", ctypes.c_uint8),
        ("_pad", ctypes.c_uint8 * 3),
    ]


class ControlPoints:
    points: list[ControlPoint]
    intensity: Intensity

    def __init__(self, points: list[ControlPoint] | list[ArrayLike], intensity: Intensity | None = None) -> None:
        match points[0]:
            case ControlPoint():
                self.points = points  # type: ignore[assignment]
            case _:
                self.points = [ControlPoint(point=p) for p in points]
        self.intensity = intensity or Intensity.MAX



import ctypes
from collections.abc import Callable
from typing import Generic, Self, TypeVar

import numpy as np

from pyautd3.autd_error import InvalidDatagramTypeError
from pyautd3.driver.datagram.datagram import Datagram
from pyautd3.driver.geometry import Device, Geometry
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import DatagramPtr, GeometryPtr

K = TypeVar("K")


class Group(Datagram, Generic[K]):
    def __init__(
        self: Self,
        key_map: Callable[[Device], K | None],
        data_map: dict[K, Datagram | tuple[Datagram, Datagram]],
    ) -> None:
        super().__init__()

        self._key_map = key_map
        self._data_map = data_map

    def _datagram_ptr(self: Self, geometry: Geometry) -> DatagramPtr:
        keymap: dict[K, int] = {}
        datagrams: np.ndarray = np.ndarray(len(self._data_map), dtype=DatagramPtr)
        keys = np.arange(len(self._data_map), dtype=np.int32)

        for k, (key, d) in enumerate(self._data_map.items()):
            match d:
                case Datagram():
                    ptr = d._datagram_ptr(geometry)
                    datagrams[k]["value"] = ptr.value
                case (Datagram(), Datagram()):
                    (d1, d2) = d
                    ptr = Base().datagram_tuple(
                        d1._datagram_ptr(geometry),
                        d2._datagram_ptr(geometry),
                    )
                    datagrams[k]["value"] = ptr.value
                case _:
                    raise InvalidDatagramTypeError
            keymap[key] = k

        def f_native(_context: ctypes.c_void_p, geometry_ptr: GeometryPtr, dev_idx: int) -> int:
            dev = Device(dev_idx, geometry_ptr)
            key = self._key_map(dev)
            return keymap[key] if key is not None else -1

        self.f_native_ = ctypes.CFUNCTYPE(ctypes.c_int32, ctypes.c_void_p, GeometryPtr, ctypes.c_uint16)(f_native)

        return Base().datagram_group(
            self.f_native_,  # type: ignore[arg-type]
            ctypes.c_void_p(0),
            geometry._geometry_ptr,
            keys.ctypes.data_as(ctypes.POINTER(ctypes.c_int32)),  # type: ignore[arg-type]
            datagrams.ctypes.data_as(ctypes.POINTER(DatagramPtr)),  # type: ignore[arg-type]
            len(keys),
        )
